/* DYNAMIC SC DENSE CODE. count-bytes version -- 
A dynamic word-based byte oriented compressor for text files based on 
(s,c)-Dense Code.

Programmed by Antonio Faria
Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "decompressor.h"
#include "version.h"

/*------------------------------------------------------------------
 reads the vocabulary size and the size of the vector "top" from 
 the aulixiar file (if it exists)
 ---------------------------------------------------------------- */
void readSizeVocabularyandTop(char *filename,  long *vocSize,  long *topSize) {
	char fileNametmp[24];
	int file;
	unsigned long value;
	
	strcpy (fileNametmp,filename);
	strcat(fileNametmp,".voc");	
	
	if ( (file = open (fileNametmp, O_RDONLY)) ==-1) {
		fprintf(stderr,"\nError while opening file: %s\n\tUsing default vocabulary size instead\n",fileNametmp);
		return ;
	}
	read(file,  vocSize,  4);
	read(file,  topSize,  4);
	
	close (file);
//	*vocSize = 885630;
//	*topSize= 9485287;
}



/*------------------------------------------------------------------
 Initializes data structures
 ---------------------------------------------------------------- */
void initialize (unsigned long sizeVoc, unsigned long tamTop){
	unsigned long i;

	_memMgr = createMemoryManager(); //_memMgr.initialize();

	N=sizeVoc;
	words  = (t_word *) malloc(N * sizeof(t_word));
	zeroNode = 0;

	topBlock = (unsigned long *) malloc(tamTop * sizeof(unsigned long));

	//initializes top and last values for each block
	MAXFREQ = 1;
	topBlock[0] =zeroNode;

	S_value = 256;
 	C_value = B_value - S_value;

	// T[0] = 0 ;	T[1] = 0 ;	T[2] = 0 ;
	TCs0 = 0 ;
	TCs1 = 0 ;
	TCs2 = 0 ;	
}


/*------------------------------------------------------------------
 * oBytes: Returns the number of bytes needed to encode the word
 * ranked "i" when using s_i value as "S"
------------------------------------------------------------------ */
unsigned long oBytes (unsigned long s_i, unsigned long i_val) {
	unsigned long k, c_i;
	unsigned long last,pot;

	//fprintf(stdFile,"\nEntering oBytes con s = %lu",s_i);
	//counts the number of bytes of the codeword to be generated
	k  	= 1;
	last  = pot = s_i;
	c_i	= B_value - s_i;

	while (last <= i_val) {
		pot *= c_i;
		last += pot;
		k++;
	}
	////fprintf(stdFile,"\n[ZERONODE=%lu]Size of i_val [%lu] when s= [%lu] is [%lu] \n",zeroNode,i_val,s_i,k);
	return k;
}


/*------------------------------------------------------------------
 * takeAccount: Adds to TCS_{s-1}, TCS_{s}, TCS_{s+1} , the number of
 * bytes needed to encode the word ranked "j", depending on the "s_x" used
------------------------------------------------------------------ */
 void takeAccount () {

	TCs0 +=oBytes (S_value - 1,j);
	TCs1 +=K; //oBytes (S_value,j);

	if (TCs0  <  TCs1) {
		S_value --;
		C_value  ++;
		//printf ("CHANGE IN S VALUE: GOIND DOWN FROM [%lu] TO [%lu]\n",S_value+1,S_value);
		//printf ("zeroNode value is [%lu]",zeroNode);
		//	fprintf(stdFile,"corpus(%lu)=%lu;\n",zeroNode,S_value);
		TCs2=TCs1;
		TCs1=TCs0;
	}	
	else {
		TCs2 += oBytes (S_value +1,j);
		if (TCs2  <  TCs1 )  {
			S_value ++;
			C_value  --;
			//printf ("CHANGE IN S VALUE: GOIND UP FROM [%lu] TO [%lu]\n",S_value+1,S_value);
			//printf ("zeroNode value is [%lu]",zeroNode);
			//	fprintf(stdFile,"corpus(%lu)=%lu;\n",zeroNode,S_value);
			TCs0=TCs1;
			TCs1=TCs2;		
		}
	}	
}



/*------------------------------------------------------------------
 * Performs the full decoding process over the compressed file.
------------------------------------------------------------------ */
void mainDecompressor (char* infile, char* outfile, int needToReadFile, unsigned long N_val, unsigned long N_val2) {
  
  unsigned char *palabra;
  unsigned long codeSize1;
  unsigned char *tmpW;
  register unsigned char abyte;
  BufferRead B;

	int f;
	/**/unsigned char *text,*top, *nextbuf;
    /**/unsigned char *pbeg,*pend,*codeStart;

	if (needToReadFile)
		readSizeVocabularyandTop(infile,(long *)&N_val,(long *)&N_val2);
	printf("\n VOC SIZE [%ld],     TOP SIZE [%ld] \n",N_val,N_val2);


	// Inicialization of read buffer
	f = open (infile, O_RDONLY);
	if (f == -1)
		printf("Cannot read file %s",*infile);
	B = bufCreate(MAX_SIZE_OF_WORD+5);    //nmero de "bytes" que hay que dejar al final del buffer siempre, por culpa
													  //de que pudiese venir una nueva palabra"que fuese cortada".
	bufSetFile (B,f);

	// Inicialization of out buffer
	outWriter = createWriter (outfile);
	
	initialize(N_val,N_val2); //initializes decompressor for up to N_value words
	
	fprintf(stderr, "\nSTARTING DECOMPRESSION...[%ld]", B->fpos);


	/************************************************************************/
	/************************************************************************/
	/*    main 1 ...                                                        */
	/************************************************************************/
	/************************************************************************/

	if (bufEmpty(B) ) { return;}

	bufCurrent (B,(char**)&text,(char**)&top);
	pbeg = text; pend = top;
    //fprintf(stderr,"\n top = [%ld]",((unsigned long) top) - ((unsigned long)B->data));

	while (( pbeg < pend )&& (zeroNode < S_value) )   {
		codeStart = pbeg;

		K=1;
		if ( (*pbeg) >= S_value ) {
			j = (*pbeg) - S_value;
			pbeg++;
			K++;

			while ((*pbeg) >= S_value) {

				j = j * C_value + ((*pbeg) - S_value);
				pbeg++;
				K++;
			}

			j = j * S_value + (*pbeg);
			if (K!=0)
				j += W[S_value][K-2];

		}
		else {
			j = (*pbeg);
		}

		//fprintf(stderr,"\n decoding=[%ld], bytes [%ld], when zeronodoe = [%ld]",j,pbeg - codeStart +1,zeroNode);

		pbeg++;

		if (j<zeroNode) {
			//	writes a word
			writeWord(outWriter,words[j].word, words[j].size);
		}
		else { // j== zeronode
			codeStart = pbeg;
			while (*pbeg !='\0') {
				pbeg++;
			}
			//	insertElement ((unsigned char *) codeStart, pbeg-codeStart, j);

			{
			register unsigned long len;
			unsigned char *w;
			
			len = pbeg-codeStart;
			
			////fprintf(stderr,"\nzeroNode= true wordLeng=[%ld]",pbeg-codeStart);
			writeWord(outWriter,codeStart, len);
			
			getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);
			
			w=words[j].word;
			register int i;
			for (i=0;i<len;i++) {  //copia tambien el '\0'
				*(w++)=*(codeStart++);
			}
			*w='\0';
			words[j].weight = 0;
			words[j].size=len;
			}

			pbeg++;
			zeroNode++;

		}/*else == zeronode*/

		//Dynamic update of the data structures of the decompressor//
		{				
			register unsigned long f,p,i,h;
			register unsigned long temp;
			unsigned char *tmpW;

			p=j;
			f = words[p].weight;
			words[p].weight +=1;

			i=j;
			j = topBlock[f];

			temp = words[i].weight;
			words[i].weight = words[j].weight;
			words[j].weight = temp;

			tmpW = words[i].word;
			words[i].word = words[j].word;
			words[j].word = tmpW;

			temp = words[i].size;
			words[i].size = words[j].size;
			words[j].size = temp;

			topBlock[f]=j+1;

			if (MAXFREQ == f+1) {  //The most frequent words need to be increased.
				topBlock[f+1] = 0;
				MAXFREQ++;
			}
			j=i;
		}//End of dynamic update
 	}/* inner while*/

	S_value --;
  	C_value  ++;
	//printf ("CHANGE IN S VALUE: GOIND DOWN FROM [%lu] TO [%lu]\n",S_value+1,S_value);
	//printf ("zeroNode value is [%lu]",zeroNode);


	/************************************************************************/
	/************************************************************************/
	/*    main 2 ...                                                        */
	/************************************************************************/
	/************************************************************************/


	while (( pbeg < pend ) && ( S_value > 210))  {
		codeStart = pbeg;

		K=1;
		if ( (*pbeg) >= S_value ) {
			j = (*pbeg) - S_value;
			pbeg++;
			K++;

			while ((*pbeg) >= S_value) {

				j = j * C_value + ((*pbeg) - S_value);
				pbeg++;
				K++;
			}

			j = j * S_value + (*pbeg);
			if (K!=0)
				j += W[S_value][K-2];

		}
		else {
			j = (*pbeg);
		}

		//fprintf(stderr,"\n decoding=[%ld], bytes [%ld], when zeronodoe = [%ld]",j,pbeg - codeStart +1,zeroNode);

		pbeg++;

		if (j<zeroNode) {
			//	writes a word
			writeWord(outWriter,words[j].word, words[j].size);
		}
		else { // j== zeronode
			codeStart = pbeg;
			while (*pbeg !='\0') {
				pbeg++;
			}
			//	insertElement ((unsigned char *) codeStart, pbeg-codeStart, j);

			{
			register unsigned long len;
			unsigned char *w;
			
			len = pbeg-codeStart;
			
			////fprintf(stderr,"\nzeroNode= true wordLeng=[%ld]",pbeg-codeStart);
			writeWord(outWriter,codeStart, len);
			
			getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);
			
			w=words[j].word;
			register int i;
			for (i=0;i<len;i++) {  //copia tambien el '\0'
				*(w++)=*(codeStart++);
			}
			*w='\0';
			words[j].weight = 0;
			words[j].size=len;
			}

			pbeg++;
			zeroNode++;

		}/*else == zeronode*/

		//Dynamic update of the data structures of the decompressor//
		{				
			register unsigned long f,p,i,h;
			register unsigned long temp;
			unsigned char *tmpW;

			p=j;
			f = words[p].weight;
			words[p].weight +=1;

			i=j;
			j = topBlock[f];

			temp = words[i].weight;
			words[i].weight = words[j].weight;
			words[j].weight = temp;

			tmpW = words[i].word;
			words[i].word = words[j].word;
			words[j].word = tmpW;

			temp = words[i].size;
			words[i].size = words[j].size;
			words[j].size = temp;

			topBlock[f]=j+1;

			if (MAXFREQ == f+1) {  //The most frequent words need to be increased.
				topBlock[f+1] = 0;
				MAXFREQ++;
			}
			j=i;
		}//End of dynamic update

		/*********************************/

		TCs0 +=oBytes(S_value -1, j);
		TCs1 +=K;  //oBytes(S_value, j);
		
		if (TCs0 < TCs1){
			S_value --;
		  	C_value ++;
		  	TCs1 = TCs0;
			//printf ("CHANGE IN S VALUE: GOIND DOWN FROM [%lu] TO [%lu]\n",S_value+1,S_value);
			//printf ("zeroNode value is [%lu]",zeroNode);
			//	fprintf(stdFile,"corpus(%lu)=%lu;\n",zeroNode,S_value);
		}
    }/* inner while*/

   	TCs2 = TCs1;

	/************************************************************************/
	/************************************************************************/
	/*    main 3 ...                                                        */
	/************************************************************************/
	/************************************************************************/

	//fprintf(stderr, "\n\n\nEntering normal process...");
	/***************************************************************************/
	/*  main 3.1, first iteration to end the pending buffer  (the first buffer)*/
	/***************************************************************************/
	while ( pbeg < pend )   {
		codeStart = pbeg;

		K=1;
		if ( (*pbeg) >= S_value ) {
			j = (*pbeg) - S_value;
			pbeg++;
			K++;

			while ((*pbeg) >= S_value) {

				j = j * C_value + ((*pbeg) - S_value);
				pbeg++;
				K++;
			}

			j = j * S_value + (*pbeg);
			if (K!=0)
				j += W[S_value][K-2];

		}
		else {
			j = (*pbeg);
		}

		//fprintf(stderr,"\n decoding=[%ld], bytes [%ld], when zeronodoe = [%ld]",j,pbeg - codeStart +1,zeroNode);

		pbeg++;

		if (j<zeroNode) {
			//	writes a word
			writeWord(outWriter,words[j].word, words[j].size);
		}
		else { // j== zeronode
			codeStart = pbeg;
			while (*pbeg !='\0') {
				pbeg++;
			}
			//	insertElement ((unsigned char *) codeStart, pbeg-codeStart, j);

			{
			register unsigned long len;
			unsigned char *w;
			
			len = pbeg-codeStart;
			
			////fprintf(stderr,"\nzeroNode= true wordLeng=[%ld]",pbeg-codeStart);
			writeWord(outWriter,codeStart, len);
			
			getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);
			
			w=words[j].word;
			register int i;
			for (i=0;i<len;i++) {  //copia tambien el '\0'
				*(w++)=*(codeStart++);
			}
			*w='\0';
			words[j].weight = 0;
			words[j].size=len;
			}

			pbeg++;
			zeroNode++;

		}/*else == zeronode*/

		//Dynamic update of the data structures of the decompressor//
		{				
			register unsigned long f,p,i,h;
			register unsigned long temp;
			unsigned char *tmpW;

			p=j;
			f = words[p].weight;
			words[p].weight +=1;

			i=j;
			j = topBlock[f];

			temp = words[i].weight;
			words[i].weight = words[j].weight;
			words[j].weight = temp;

			tmpW = words[i].word;
			words[i].word = words[j].word;
			words[j].word = tmpW;

			temp = words[i].size;
			words[i].size = words[j].size;
			words[j].size = temp;

			topBlock[f]=j+1;

			if (MAXFREQ == f+1) {  //The most frequent words need to be increased.
				topBlock[f+1] = 0;
				MAXFREQ++;
			}
			j=i;
		}//End of dynamic update

		takeAccount();

    }/* inner while*/


	nextbuf=pbeg;
	bufLoad (B, (char*)nextbuf);


	/***************************************************************************/
	/*  main 3.2, 
	/***************************************************************************/
	//fprintf(stderr,"\n MAIN 3 GENERAL");

	while (!bufEmpty(B) ) {

		//fprintf(stderr, "\nEntra en while principal");

		bufCurrent (B,(char**)&text,(char**)&top);
		pbeg = text; pend = top;
	    //fprintf(stderr,"\n top = [%ld]",((unsigned long) top) - ((unsigned long)B->data));

		while ( pbeg < pend )   {
			codeStart = pbeg;
	
			K=1;
			if ( (*pbeg) >= S_value ) {
				j = (*pbeg) - S_value;
				pbeg++;
				K++;
	
				while ((*pbeg) >= S_value) {
	
					j = j * C_value + ((*pbeg) - S_value);
					pbeg++;
					K++;
				}
	
				j = j * S_value + (*pbeg);
				if (K!=0)
					j += W[S_value][K-2];
	
			}
			else {
				j = (*pbeg);
			}
	
			//fprintf(stderr,"\n decoding=[%ld], bytes [%ld], when zeronodoe = [%ld]",j,pbeg - codeStart +1,zeroNode);
	
			pbeg++;
	
			if (j<zeroNode) {
				//	writes a word
				writeWord(outWriter,words[j].word, words[j].size);
			}
			else { // j== zeronode
				codeStart = pbeg;
				while (*pbeg !='\0') {
					pbeg++;
				}
				//	insertElement ((unsigned char *) codeStart, pbeg-codeStart, j);
	
				{
				register unsigned long len;
				unsigned char *w;
				
				len = pbeg-codeStart;
				
				////fprintf(stderr,"\nzeroNode= true wordLeng=[%ld]",pbeg-codeStart);
				writeWord(outWriter,codeStart, len);
				
				getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);
				
				w=words[j].word;
				register int i;
				for (i=0;i<len;i++) {  //copia tambien el '\0'
					*(w++)=*(codeStart++);
				}
				*w='\0';
				words[j].weight = 0;
				words[j].size=len;
				}
	
				pbeg++;
				zeroNode++;
	
			}/*else == zeronode*/
	
			//Dynamic update of the data structures of the decompressor//
			{				
				register unsigned long f,p,i,h;
				register unsigned long temp;
				unsigned char *tmpW;
	
				p=j;
				f = words[p].weight;
				words[p].weight +=1;
	
				i=j;
				j = topBlock[f];
	
				temp = words[i].weight;
				words[i].weight = words[j].weight;
				words[j].weight = temp;
	
				tmpW = words[i].word;
				words[i].word = words[j].word;
				words[j].word = tmpW;
	
				temp = words[i].size;
				words[i].size = words[j].size;
				words[j].size = temp;
	
				topBlock[f]=j+1;
	
				if (MAXFREQ == f+1) {  //The most frequent words need to be increased.
					topBlock[f+1] = 0;
					MAXFREQ++;
				}
				j=i;
			}//End of dynamic update
	
			takeAccount();

		}/* inner while*/

		nextbuf=pbeg;
		bufLoad (B, (char*)nextbuf);

	}/* outter while */

	
	fprintf(stderr, "\n... DECOMPRESSION FINISHED \n\n");

	bufDestroy(B);
	destroyWriter(outWriter);
	free(words);

} //mainDecompressor



/*------------------------------------------------------------------ */
/*
 *  DECOMPRESSOR PROGRAM OF DYNAMIC (S,C)-DC.
 */
/*------------------------------------------------------------------ */
int main(int argc, char* argv[])
{
	char fileName[20];

	long sizeNValue;
	long sizeNValue2;
	int needToLoadSizes=0;
	
	showMessage();
	if (argc < 3) {
		fprintf(stderr,"\n Wrong call to DSCDC dynamic decompressor\n");
		fprintf(stderr," Use:  %s  <in file>  <out file> [<VOC SIZE]><TOP SIZE]>\n",argv[0]);
		return 0;
	}

	sizeNValue = MAX_LEN_VOCABULARY;
	sizeNValue2 = INITIAL_TOP_FREQ;
	
    if ((argc ==3) || (argc = 4)) 
    	needToLoadSizes = 1;
	
	if (argc ==5) {
		sizeNValue = atol(argv[3]);
		sizeNValue2 = atol(argv[4]);
		needToLoadSizes = 0;
	}


	StartRusageTime();
	mainDecompressor(argv[1], argv[2],needToLoadSizes,sizeNValue,sizeNValue2);
	StopRusageTimeFile("timesDecDSCDC.txt",argv[1]);
	fprintf(stderr,"\n ZERONODE VALUE AT THE END IS:   %ld \n",zeroNode);
}
